from .branin import Branin
from .cantilever_beam import CantileverBeam
from .sphere import Sphere
from .reduced_problem import ReducedProblem
from .robot_arm import RobotArm
from .rosenbrock import Rosenbrock
from .tensor_product import TensorProduct
from .torsion_vibration import TorsionVibration
from .water_flow import WaterFlow
from .water_flow_lfidelity import WaterFlowLFidelity
from .welded_beam import WeldedBeam
from .wing_weight import WingWeight
from .ndim_cantilever_beam import NdimCantileverBeam
from .ndim_robot_arm import NdimRobotArm
from .ndim_rosenbrock import NdimRosenbrock
from .ndim_step_function import NdimStepFunction
from .lp_norm import LpNorm
